import ujson

from ..types import FormatTemplate


class BlobMeta:
    def __init__(self):
        pass

    def __str__(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def serialize(self):
        """
        :rtype: dict
        """
        raise NotImplementedError


class JsonBlobMeta(BlobMeta):
    def __init__(self, jnode):
        """
        :type jnode: dict
        """
        BlobMeta.__init__(self)
        self._jnode = jnode

    def __str__(self):
        """
        :rtype: str
        """
        return ujson.dumps(self._jnode)

    def serialize(self):
        """
        :rtype: dict
        """
        return self._jnode


class Blob:
    """
    :type _name: str | None
    """

    def __init__(self, profile, file_ext):
        """
        :type profile: str
        :type file_ext: str
        """
        self._name = None
        self._profile = profile
        self._file_ext = file_ext

    @property
    def name(self):
        """
        :rtype: str
        """
        if self._name is None:
            raise RuntimeError

        return self._name

    @name.setter
    def name(self, value):
        """
        :type value: str
        """
        if len(value):
            self._name = value
        else:
            raise RuntimeError

    @property
    def profile(self):
        """
        :rtype: str
        """
        return self._profile

    @property
    def min_ts(self):
        """
        :return: value in usec
        :rtype: int
        """
        raise NotImplementedError

    @property
    def max_ts(self):
        """
        :return: value in usec
        :rtype: int
        """
        raise NotImplementedError

    @property
    def file_path(self):
        """
        :rtype: str | None
        """
        raise NotImplementedError

    @property
    def file_ext(self):
        """
        :rtype: str
        """
        return self._file_ext

    @property
    def file_data(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def file_meta(self):
        """
        :rtype: str | None
        """
        return None

    def serialize(self, tz_service=None):
        """
        :type tz_service: ITzService | None
        :rtype: dict
        """
        raise NotImplementedError


class MemBlob(Blob):
    def __init__(self, profile, ts, data, meta=None, file_ext='jpeg'):
        """
        :type ts: int
        :type profile: str
        :type data: str
        :type meta: BlobMeta | None
        """
        Blob.__init__(self, profile, file_ext=file_ext)

        self._ts = ts
        self._data = data
        self._meta = meta

    @property
    def min_ts(self):
        """
        :rtype: int
        """
        return self._ts

    @property
    def max_ts(self):
        """
        :rtype: int
        """
        return self._ts

    @property
    def file_path(self):
        """
        :rtype: str | None
        """
        return None

    @property
    def file_data(self):
        """
        :rtype: str
        """
        return self._data

    def serialize(self, tz_service=None):
        """
        :type tz_service: ITzService | None
        :rtype: dict
        """
        # jnode = {
        #     "name": "0_150402739915.jpeg",
        #     "info": {
        #         "size": [1920, 1200, 100500],
        #         "trans": {
        #             "o": [0.0, 0.0],
        #             "i": [1.0, 0.0],
        #             "j": [0.0, 1.0]
        #         }
        #     }
        # }

        jnode = {
            'name': self.name,
        }

        if tz_service:
            jnode['time'] = tz_service.get_datetime_isostr(self.min_ts / 1e6)

        if self._meta:
            jnode['info'] = self._meta.serialize()

        return jnode


class FileBlob(Blob):
    """
    :type _file: file
    """
    def __init__(self, min_ts, max_ts, data, meta=None, profile='measurement'):
        """
        :type min_ts: int
        :type max_ts: int
        :type data: file
        :type meta: BlobMeta | None
        """
        Blob.__init__(self, profile, file_ext='mp4')

        self._min_ts = min_ts
        self._max_ts = max_ts

        self._file = data
        self._meta = meta

    @property
    def min_ts(self):
        """
        :rtype: int
        """
        return self._min_ts

    @property
    def max_ts(self):
        """
        :rtype: int
        """
        return self._max_ts

    @property
    def file_path(self):
        """
        :rtype: str
        """
        return self._file.name

    @property
    def file_data(self):
        """
        :rtype: str
        """
        return self._file.read()

    def serialize(self, tz_service=None):
        """
        :type tz_service: ITzService | None
        :rtype: dict
        """
        jnode = {
            'name': self.name,
        }

        if tz_service:
            jnode['time'] = [
                tz_service.get_datetime_isostr(self.min_ts / 1e6),
                tz_service.get_datetime_isostr(self.max_ts / 1e6)
            ]

        if self._meta:
            jnode['info'] = self._meta.serialize()

        return jnode


class PlateImArgs:
    def __init__(self):
        self.width, self.height = 134, 30  # ~ 5:1


class ThumbImArgs:
    def __init__(self):
        # self.width, self.height = 220, 165  # 4:3
        self.width, self.height = 224, 224  # 1:1 for MMR


class VehicleImAnnoArgs:
    def __init__(self, plate=False):
        """
        :type plate: bool
        """
        self.plate = plate


class VehicleImArgs:
    def __init__(self, width, footer, anno):
        """
        :type width: int
        :type footer: ImFooterArgs
        :type anno: VehicleImAnnoArgs
        """
        self.wh_ratio = 4.0 / 3.0

        self.width = width
        self.height = width / self.wh_ratio  # 4:3

        self.footer = footer
        self.anno = anno


class SceneImAnnoArgs:
    def __init__(self, track=False, vehicle=False, plate=False):
        """
        :type track: bool
        :type vehicle: bool
        :type plate: bool
        """
        self.track = track
        self.vehicle = vehicle
        self.plate = plate


class SceneImArgs:
    def __init__(self, width, footer, anno):
        """
        :type width: int
        :type footer: ImFooterArgs
        :type anno: SceneImAnnoArgs
        """
        self.width = width
        self.footer = footer
        self.anno = anno


class LiveImArgs:
    def __init__(self, footer):
        """
        :type footer: ImFooterArgs
        """
        self.footer = footer


class ImFooterColArgs:
    def __init__(self, width, align, template):
        """
        :type width: int
        :type align:  int
        :type template: str
        """
        self.width = width
        self.align = align
        self.template = template

    def eval(self, mapping):
        """
        :type mapping: dict[str, str]
        :rtype: ImFooterColArgs
        """
        value = FormatTemplate(self.template).substitute(mapping)
        return ImFooterColArgs(self.width, self.align, value)


class ImFooterRowArgs:
    def __init__(self, enabled, cols):
        """
        :type enabled: bool
        :type cols: list of ImFooterColArgs
        """
        self.enabled = enabled
        self.cols = cols

    def eval(self, mapping):
        """
        :type mapping: dict[str, str]
        :rtype: ImFooterRowArgs
        """
        cols = map(lambda col: col.eval(mapping), self.cols)
        return ImFooterRowArgs(self.enabled, cols)


class ImFooterArgs:
    def __init__(self, height, font_size, rows):
        """
        :param int height:
        :param int font_size: Font size, pixels
        :param list[ImFooterRowArgs] rows:
        """
        self.height = height
        self.font_size = font_size
        self.rows = rows

    def eval(self, footer_vars):
        """
        :type footer_vars: ImFooterVars
        :rtype: ImFooterArgs
        """
        var_names = set()
        for row in self.rows:
            for col in row.cols:
                var_names.update(FormatTemplate.get_pholders(col.template))

        mapping = {}
        for var_name in var_names:
            if var_name == 'datetime':
                mapping[var_name] = footer_vars.datetime

            elif var_name == 'datetime-ms':
                mapping[var_name] = footer_vars.datetime_ms

            elif var_name == 'post-name':
                mapping[var_name] = footer_vars.location

            elif var_name == 'geolocation':
                mapping[var_name] = footer_vars.geolocation

            elif var_name == 'plate':
                mapping[var_name] = footer_vars.plate_text

            elif var_name == 'speed':
                mapping[var_name] = footer_vars.spot_speed

            elif var_name == 'serial':
                mapping[var_name] = footer_vars.serial

            else:
                raise NotImplementedError

        rows = map(lambda r: r.eval(mapping), self.rows)
        return ImFooterArgs(self.height, self.font_size, rows)


class ImFooterVars:
    def __init__(self, tz_service, ts, defs=None, location=None, geolocation=None,
                 plate_text=None, spot_speed=None, serial=''):
        """
        :type tz_service: ITzService
        :type ts: long
        :type defs: dict | None
        :type location: Location | None
        :type geolocation: Geolocation | None
        :type plate_text: str | None
        :type spot_speed: float | None
        :type serial: str
        """
        dt = tz_service.get_datetime(ts / 1e6)
        # <22.01.2018 17:00:33>
        self.datetime = dt.strftime('%d.%m.%Y %H:%M:%S')
        # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
        self.datetime_ms = dt.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]

        defs = defs or dict()
        self.location = location.place if (location is not None and location.place != '') else defs.get('post-name', '')
        self.geolocation = str(geolocation) if geolocation else defs.get('geolocation', '')
        self.plate_text = plate_text or defs.get('plate', '')
        self.spot_speed = str(spot_speed or defs.get('speed', ''))
        self.serial = serial


class ImEngineArgs:
    def __init__(self, scene, scene_footer, vehicle=None, vehicle_footer=None):
        """
        :type scene: SceneImArgs
        :type vehicle: VehicleImArgs
        :type scene_footer: dict
        :type vehicle_footer: dict
        """
        self.scene = scene
        self.vehicle = vehicle
        self.thumb = ThumbImArgs()
        self.plate = PlateImArgs()

        self.scene_footer = scene_footer
        self.body_footer = vehicle_footer
