import os

from . import eventlog
from .types import ImageRef
from .com import fs

from edge_rawutil import rawutil


class ImageFileQueue:
    def __init__(self, twisted_reactor, target_codec, root_dir, min_file_count, max_file_count, log_level='info'):
        """
        :type target_codec: str | None
        :type root_dir: str
        :type min_file_count: int
        :type max_file_count: int
        :type log_level: str
        """
        self._root_dir = root_dir
        self._min_file_count = min_file_count
        self._max_file_count = max_file_count

        self._target_codec = target_codec

        self._rawutil = rawutil.RawUtil.create(rawutil.MAX_WIDTH, rawutil.MAX_HEIGHT, log_level)
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._rawutil.destroy)

        fs.silent_mkdir_r(self._root_dir)
        file_names = os.listdir(self._root_dir)

        file_names.sort(reverse=False, key=self._unpack_ts)

        self._last_file_suffix = self._unpack_ts(file_names.pop()) if file_names else 0
        self._expected_file_count = len(file_names)

    @staticmethod
    def _unpack_ts(name):
        """
        :type name: str
        :rtype: int
        """
        return int(name.split('-')[-1])

    @classmethod
    def _cleanup(cls, root_dir, min_file_count):
        """
        :type root_dir: str
        :type min_file_count: int
        :rtype: int
        """

        file_names = os.listdir(root_dir)
        file_names.sort(reverse=False, key=cls._unpack_ts)

        actual_file_count = len(file_names)
        for file_name in file_names:
            if actual_file_count <= min_file_count:
                break

            fs.silent_rm_r(fs.join_path(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self, next_name):
        """
        :type next_name: int
        :rtype: str
        """
        self._last_file_suffix += 1
        next_path = fs.join_path(self._root_dir, '{}-{}'.format(self._last_file_suffix, next_name))
        return next_path

    def add_image_ref(self, image_ref):
        """
        :type image_ref: ImageRef
        :rtype: ImageRef
        """
        next_filepath = self._get_next_filepath(image_ref.ts)

        if self._target_codec is None:
            fs.edit_dir(self._root_dir, os.link, image_ref.url, next_filepath)

        elif image_ref.codec == 'JPEG' and self._target_codec == 'JPEG':
            fs.edit_dir(self._root_dir, os.link, image_ref.url, next_filepath)

        else:
            rgba_pil_image = self._rawutil.handle_image(
                image_ref.url, image_ref.codec, image_ref.width, image_ref.height)
            if not rgba_pil_image:
                raise RuntimeError('convert raw-image to rgba')

            if self._target_codec == 'JPEG':
                fs.edit_dir(self._root_dir, fs.write_jpeg_file, rgba_pil_image, next_filepath)

            elif self._target_codec == 'RGBA':
                fs.edit_dir(self._root_dir, fs.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=image_ref.ts, url=next_filepath, codec=self._target_codec, width=image_ref.width,
                        height=image_ref.height)

    def add_image(self, ts, rgba_pil_image):
        """
        :type ts: int | None
        :type rgba_pil_image: PIL.Image.Image
        :rtype: ImageRef
        """
        next_filepath = self._get_next_filepath(ts)

        if self._target_codec == 'JPEG':
            fs.edit_dir(self._root_dir, fs.write_jpeg_file, rgba_pil_image, next_filepath)

        elif self._target_codec == 'RGBA':
            fs.edit_dir(self._root_dir, fs.write_rgba_file, rgba_pil_image, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        w, h = rgba_pil_image.size
        return ImageRef(ts=ts, url=next_filepath, codec=self._target_codec, width=w, height=h)

    def add_blob_image(self, blob_bytes, blob_ts=None, blob_codec=None):
        """
        :type blob_bytes: str
        :type blob_ts: int | None
        :type blob_codec: str | None
        :rtype: ImageRef
        """
        next_filepath = self._get_next_filepath(blob_ts)

        # write file to the queue
        def write_bin_file(content_bytes):
            with open(next_filepath, 'wb') as bin_file:
                bin_file.write(content_bytes)

        fs.edit_dir(self._root_dir, write_bin_file, blob_bytes)

        # cleanup the queue
        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)

        return ImageRef(ts=blob_ts, url=next_filepath, codec=blob_codec)

    def remove_image(self, image_ref):
        """
        :type image_ref: LrImageRef
        """
        # assert image_ref.codec == 'JPEG' #todo : move to where called

        # noinspection PyBroadException
        try:
            os.unlink(image_ref.url)
            self._expected_file_count -= 1
        except Exception:
            eventlog.log_x()


class JsonFileQueue:
    def __init__(self, args):
        """
        :type args: QueueArgs
        """
        self._root_dir = args.endpoint
        self._min_file_count = args.min_file_count
        self._max_file_count = args.max_file_count

        fs.silent_mkdir_r(self._root_dir)
        file_name_list = map(str, os.listdir(self._root_dir))
        file_name_list.sort(key=lambda name: int(name.split('-')[0]))

        self._last_file_suffix = int(file_name_list.pop().split('-')[0]) if file_name_list else 0
        self._expected_file_count = len(file_name_list)

    @staticmethod
    def _cleanup(root_dir, min_file_count):
        """
        :type root_dir: str
        :type min_file_count: int
        :rtype: int
        """
        file_name_list = map(str, os.listdir(root_dir))
        file_name_list.sort(key=lambda name: int(name.split('-')[0]))

        actual_file_count = len(file_name_list)
        for file_name in file_name_list:
            if actual_file_count <= min_file_count:
                break

            fs.silent_rm_r(fs.join_path(root_dir, str(file_name)))
            actual_file_count -= 1

        return actual_file_count

    def _get_next_filepath(self, next_filename=None):
        """
        :type next_filename: str | None
        :rtype: str
        """
        self._last_file_suffix += 1
        if next_filename:
            next_filepath = fs.join_path(self._root_dir, '%s-%s' % (self._last_file_suffix, next_filename))
        else:
            next_filepath = fs.join_path(self._root_dir, str(self._last_file_suffix))

        return next_filepath

    def add_jnode(self, jnode, next_filename=None):
        """
        :type jnode: dict
        :type next_filename: str | None
        """
        next_filepath = self._get_next_filepath(next_filename)
        fs.edit_dir(self._root_dir, fs.write_json_file, jnode, next_filepath)

        self._expected_file_count += 1
        if self._expected_file_count >= self._max_file_count:
            self._expected_file_count = self._cleanup(self._root_dir, self._min_file_count)
